
//*****************************************************************
//***
//***   File:     VECTORH.C++
//***   Function: Implementating VectorH and derived classes
//***
//***   Compilation: Ok.
//***
//***   11-Dic-95        Carlos        Creation
//***   20-May-96        Jorge         Added methods in Vector3D
//***                                  (Module, Normalized, and
//***                                   RandPointOnSphere).
//***   21-May-96        Jorge         Modification RandPointOnSphere.
//***   26-Nov-96        Carlos        Added 'Vector3D::ReflectedOn' 
//***                                  and 'Vector3D::operator=='
//***   21-Jan-97        Jorge         Added 'Vector3D::RefractedOn'
//***   15-Jul-97	 Jorge	       Modified 'Vector3D::operator *' and
//***                                  'Point3D::operator *' for Linux-GNU
//***                                  compiling with free errors.
//***
//*****************************************************************

#include <Global.h++>
#include <VectorH.h++>
#include <Transform.h++>
#include <Point2D.h++>

#include <math.h>

#define impl_un(o)   for( int i =0 ; i<4 ; i++ )  c[i] o v[(Axis)i] 

//***************************************************
//***
//***   VectorH
//***
//***************************************************

VectorH::VectorH( Scalar cx, Scalar cy, Scalar cz, Scalar cw ) 
{
  c[x]= cx ;
  c[y]= cy ; 
  c[z]= cz ;
  c[w]= cw ;
}



Scalar VectorH::operator[]( const Axis axis ) const
{
  return c[(int)axis] ;
}

VectorH & VectorH::operator *= ( const Matrix4x4 &m )
{
   Scalar res[4] ;
   
   int i ;

   for( i= 0 ; i<4 ; i++ )
   {  Scalar tmp= 0.0 ;
      for( int j= 0 ; j<4 ; j++ )
         tmp += c[j]*m.a[j][i] ;
      res[i]= tmp ;
   }
   
   for( i= 0 ; i<4 ; i++ )
     c[i]= res[i] ;
     
   return *this ;  

}

VectorH & VectorH::operator *= ( const Transform &t )
{
   return (*this) *= t.mDir ;
}

VectorH VectorH::operator * ( const Transform &t ) const 
{
   VectorH res= *this ;
   return res *= t ;
}

Boolean VectorH::IsValid() const

{
   Axis i ;

   FOR_ALL_AXIS(i)
   {
      if ( !IsNumber( c[i] ) )
         return false ;
   }

   return true ;

}

ostream & operator << ( ostream & os, const VectorH & v ) 
{
   return os << "(" << v[x] 
             << "," << v[y] 
             << "," << v[z] 
             << "," << v[w] 
             << ")" ;
}

/*************************
***
***   Vector3D
***
**************************/

Vector3D::Vector3D( Scalar cx, Scalar cy, Scalar cz ) 

: VectorH(cx,cy,cz,0.0)

{ }


//*************************************************************

Vector3D::Vector3D
( 
  Axis axis 
) 

: VectorH(0,0,0,0)
{
  c[axis]= 1.0 ;
}

//*************************************************************

Vector3D & Vector3D::operator += ( const Vector3D &v )
{
   impl_un(+=) ;
   return *this ;
}

Vector3D Vector3D::operator + ( const Vector3D &v ) const
{
  Vector3D res(*this) ;  
  return res += v ;
}

Vector3D & Vector3D::operator -= ( const Vector3D &v )
{
   impl_un(-=) ;
   return *this ;
}

Vector3D Vector3D::operator - ( const Vector3D &v ) const
{
  Vector3D res(*this) ;  
  return res -= v ;
}

Scalar Vector3D::operator | ( const Vector3D &v ) const
{
   return c[x]*v[x] + c[y]*v[y] + c[z]*v[z] ;
}

Vector3D & Vector3D::operator *= ( Scalar t )
{
   c[x] *= t ;
   c[y] *= t ; 
   c[z] *= t ;
   
   return *this ;
}

Vector3D Vector3D::operator * ( Scalar t ) const 
{
   Vector3D res(*this) ;   
   return res *= t ;  
}

Vector3D operator * ( Scalar t, const Vector3D &v )
{
   Vector3D res(v) ;  
   return res *= t ;
}

Vector3D Vector3D::operator * ( const Vector3D &v ) const
{
   return Vector3D( c[y]*v[z] - c[z]*v[y],
                    c[z]*v[x] - c[x]*v[z],
                    c[x]*v[y] - c[y]*v[x] ) ;
}

Vector3D  & Vector3D::operator *= ( const Transform &t ) 
{ 
   (*((VectorH *)this)) *= t ;
   return *this ;
}

Vector3D Vector3D::operator * ( const Transform &t ) const 
{
   Vector3D res= *this ;   
   return res *= t ;
}

Scalar Vector3D::Module ( ) const 
{
   return sqrt ( sqr(c[x]) + sqr(c[y]) + sqr(c[z]) ) ;
}

Vector3D Vector3D::Normalized ( ) const
{
   Scalar mod= 1.0 / Module() ;
   return (*this)*mod ;
}  

Vector3D Vector3D::ReflectedOn( const Vector3D &normal ) const 
{
   Vector3D result= normal ;

   result *= ( normal | (*this) ) ;
   result -= (*this) ;
   result *= 2.0 ;
   result += (*this) ;  

   return result ;
}

Vector3D Vector3D::RefractedOn ( const Vector3D &normal, Scalar irf ) const
{
   Scalar ci= -(*this | normal) ;
   Scalar a= 1.0 + sqr(irf) * (sqr(ci)-1.0) ;
   
   if (a<0.0)
      return Vector3D (0.0, 0.0, 0.0) ;

   Scalar b= irf*ci - sqrt(a) ;
   
   return (*this) * irf + normal*b ;
}

        
Boolean Vector3D::operator==( const Vector3D &v ) const 
{
    return (Boolean)( ( c[x] == v[x] ) && ( c[y]==v[y] ) && ( c[z]==v[z] ) ) ;
}

//****************************************************************

Vector3D RandPointOnSphere ( )
{
   
   while(true)
   {
       Vector3D v( rnd(), rnd(), rnd() ) ;
       Scalar m= v.Module() ;
   
       if ( m > 0.0 && m <= 1.0 )
          return v *= (1.0/m) ;
   }
   
}

//****************************************************************

Vector3D RandVectorCosineZ ( )
{
   Point2D p= RandPointOnCircle() ;

   Scalar xx= p[u], 
          yy= p[v], 
          zz= sqrt(1.0-(p[u]*p[u])-(p[v]*p[v]) ) ;

   if ( xx != xx || yy != yy || zz != zz )
      cout << "\twarning: not a number in RandVectorCosineZ\n" 
           << "\t xx == " << xx << "   yy == " << yy << "  zz == " << zz << "\n" 
           << "\t p[u] == " << p[u] << "   p[v] == " << p[v] 
           << "\t r= " << p[u]*p[u] + p[v]*p[v] << "\n" << flush ;

   Vector3D res( xx,yy,zz ) ;
   return res ;
}

//***************************************************************

Vector3D RandVectorPhongZ( Scalar n ) 

// this returns a random vector with unit lenght.
// the probability for selecting any vector is proportional
// to inner product of the vector and (0,0,1), raised to n.

//  see  "A Ray Tracing Framework for Global Illumination Systems"
//       Shirley, Sung, Brown
//       Proceedings of "Graphics Interface'91"
//       (see bottom of page 120 and top of 121)


{
   Scalar uu= rnd() ,
          vv= rnd() ;

   Scalar beta= 0.5*M_PI - acos( pow( 1.0-uu , 1.0/(n+1.0) ) ),
          alfa= 2.0*M_PI*vv ;

   //cout << "alfa= " << alfa << "\n"
   //     << "beta= " << beta << "\n" << flush ;
   
   Vector3D res( cos(alfa)*cos(beta) , sin(alfa)*cos(beta), sin(beta) ) ;
  
   //cout << "RandVectorPhongZ returns ---> " << res << "  mod=" << res.Module() << "\n" << flush ;

   return res ;
}




/*************************
***
***   Point3D
***
**************************/

Point3D::Point3D( Scalar cx, Scalar cy, Scalar cz ) 

: VectorH(cx,cy,cz,1.0)

{  }

Vector3D Point3D::operator - ( const Point3D & p ) const 
{
  return Vector3D( c[x]-p[x], c[y]-p[y], c[z]-p[z] ) ;
}

Point3D & Point3D::operator += ( const Vector3D & v )

{
   impl_un(+=) ;
   return *this ;
}
        
Point3D Point3D::operator + ( const Vector3D & v ) const 
{
   Point3D res(*this) ;
   return res += v ;
}
        
Point3D & Point3D::operator -= ( const Vector3D & v ) 
{
   impl_un(-=) ;
   return *this ;
}
        
Point3D Point3D::operator - ( const Vector3D &v ) const 
{
    Point3D res(*this) ;
    return res -= v ;
}



Point3D  & Point3D::operator *= ( const Transform &t ) 
{ 
   (*((VectorH *)this)) *= t ;
   return *this ;
}

Point3D Point3D::operator * ( const Transform &t ) const 
{
   Point3D res= *this ;   
   return res *= t ;
}

Point3D  LinearInterpolate( Scalar t, const Point3D &p0, const Point3D &p1 )
{
  Scalar t0= 1.0-t ;

  return Point3D(
     t0*p0.c[x] + t*p1.c[x],  
     t0*p0.c[y] + t*p1.c[y], 
     t0*p0.c[z] + t*p1.c[z]   
  ) ;

}


